#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.

import asyncio
import atexit
import os
import threading

from . import consts
from . import logging as pc_logging
from . import project_config
from . import project
from . import runtime_python_all
from . import project_factory_local as rfl
from . import project_factory_git as rfg
from . import project_factory_tar as rft
from .user_config import user_config
from .utils import total_size


# Context
class Context(project_config.Configuration):
    """Stores and caches all imported objects."""

    stats_packages: int
    stats_packages_instantiated: int
    stats_parts: int
    stats_parts_instantiated: int
    stats_assemblies: int
    stats_assemblies_instantiated: int
    stats_memory: int

    def __init__(self, config_path="."):
        """Initializes the context and imports the root project."""
        super().__init__(consts.THIS, config_path)

        # Protect the critical sections from access in different threads
        self.lock = threading.Lock()

        self.option_create_dirs = False
        self.runtimes_python = {}

        self.stats_packages = 0
        self.stats_packages_instantiated = 0
        self.stats_parts = 0
        self.stats_parts_instantiated = 0
        self.stats_assemblies = 0
        self.stats_assemblies_instantiated = 0
        self.stats_memory = 0

        if os.path.isdir(config_path):
            config_path = "."
        else:
            # TODO(clairbee): Add support for config files not at the top level
            config_path = os.path.basename(config_path)

        # self.projects contains all projects known to this context
        self.projects = {}
        self._projects_being_loaded = {}

        with pc_logging.Process("ImportDeps", self.config_dir):
            self.import_project(
                None,  # parent
                {
                    "name": consts.THIS,
                    "type": "local",
                    "path": config_path,
                },
            )

    def stats_recalc(self, verbose=False):
        self.stats_memory = total_size(self, verbose)

    def get_project(self, project_name) -> project.Project:
        if not project_name in self.projects:
            pc_logging.error("The project '%s' is not found." % project_name)
            pc_logging.error("%s" % self.projects)
            return None

        config = self.projects[project_name]
        return config

    def import_project(self, parent, project_import_config):
        if not "name" in project_import_config or not "type" in project_import_config:
            pc_logging.error(
                "Invalid project configuration found: %s" % project_import_config
            )
            return None

        name = project_import_config["name"]
        with pc_logging.Action("Import", name):
            if name in self._projects_being_loaded:
                pc_logging.error(
                    "Recursive project loading detected (%s), aborting." % name
                )
                return None

            # Depending on the project type, use different factories
            if (
                not "type" in project_import_config
                or project_import_config["type"] == "local"
            ):
                rfl.ProjectFactoryLocal(self, parent, project_import_config)
            elif project_import_config["type"] == "git":
                with pc_logging.Action("Git", name):
                    rfg.ProjectFactoryGit(self, parent, project_import_config)
            elif project_import_config["type"] == "tar":
                with pc_logging.Action("Tar", name):
                    rft.ProjectFactoryTar(self, parent, project_import_config)
            else:
                pc_logging.error("Invalid project type found: %s." % name)
                return None

            # Check whether the factory was able to successfully add the project
            if not name in self.projects:
                pc_logging.error(
                    "Failed to create the project: %s" % project_import_config
                )
                return None

            self.stats_packages += 1
            self.stats_packages_instantiated += 1

            imported_project = self.projects[name]

            # Load the dependencies recursively
            # while preventing circular dependencies
            self._projects_being_loaded[name] = True
            # pc_logging.debug("Imported config: %s..." % imported_project.config_obj)
            if (
                "import" in imported_project.config_obj
                and not imported_project.config_obj["import"] is None
            ):
                for prj_name in imported_project.config_obj["import"]:
                    # pc_logging.debug("Importing: %s..." % prj_name)
                    prj_conf = imported_project.config_obj["import"][prj_name]
                    prj_conf["name"] = prj_name
                    self.import_project(imported_project, prj_conf)
            del self._projects_being_loaded[name]

            return imported_project

    def get_packages(self):
        return map(
            lambda pkg: {"name": pkg.name, "desc": pkg.desc}, self.projects.values()
        )

    def _get_part(self, part_name, project_name, params=None):
        prj = self.get_project(project_name)
        if prj is None:
            pc_logging.error("Package %s not found" % project_name)
            # Don't print anything as self.get_project is expected to report errors
            return None
        pc_logging.debug("Retriving %s from %s" % (part_name, project_name))
        return prj.get_part(part_name, params)

    def get_part(self, part_name, project_name, params=None):
        with self.lock:
            return self._get_part(part_name, project_name, params)

    def get_part_shape(self, part_name, project_name, params=None):
        with self.lock:
            return asyncio.run(
                self._get_part(part_name, project_name, params).get_shape()
            )

    def get_part_cadquery(self, part_name, project_name, params=None):
        with self.lock:
            return asyncio.run(
                self._get_part(part_name, project_name, params).get_cadquery()
            )

    def get_part_build123d(self, part_name, project_name, params=None):
        with self.lock:
            return asyncio.run(
                self._get_part(part_name, project_name, params).get_build123d()
            )

    def _get_assembly(self, assembly_name, project_name, params=None):
        prj = self.get_project(project_name)
        if prj is None:
            # Don't print anything as self.get_project is expected to report errors
            return None
        return prj.get_assembly(assembly_name, params)

    def get_assembly(self, assembly_name, project_name, params=None):
        with self.lock:
            return self._get_assembly(assembly_name, project_name, params)

    def get_assembly_shape(self, assembly_name, project_name, params=None):
        with self.lock:
            return asyncio.run(
                self._get_assembly(assembly_name, project_name, params).get_shape()
            )

    def get_assembly_cadquery(self, assembly_name, project_name, params=None):
        with self.lock:
            return asyncio.run(
                self._get_assembly(assembly_name, project_name, params).get_cadquery()
            )

    def get_assembly_build123d(self, assembly_name, project_name, params=None):
        with self.lock:
            return asyncio.run(
                self._get_assembly(assembly_name, project_name, params).get_build123d()
            )

    def render(self, format=None, output_dir=None):
        prj = self.get_project(consts.THIS)
        prj.render(format=format, output_dir=output_dir)

    def get_python_runtime(self, version, python_runtime=None):
        if python_runtime is None:
            python_runtime = user_config.python_runtime
        runtime_name = python_runtime + "-" + version
        if not runtime_name in self.runtimes_python:
            self.runtimes_python[runtime_name] = runtime_python_all.create(
                self, version, python_runtime
            )
        return self.runtimes_python[runtime_name]

    def ensure_dirs(self, path):
        if not self.option_create_dirs:
            return
        os.makedirs(path)

    def ensure_dirs_for_file(self, filename):
        if not self.option_create_dirs:
            return
        path = os.path.dirname(filename)
        os.makedirs(path)



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-26
#
# Licensed under Apache License, Version 2.0.
#

import copy
import typing

from . import part_config
from . import part_factory as pf
from . import logging as pc_logging


class PartFactoryEnrich(pf.PartFactory):
    target_part: str
    target_project: typing.Optional[str]

    def __init__(self, ctx, project, config):
        with pc_logging.Action("Enrich", project.name, config["name"]):
            super().__init__(ctx, project, config)

            # Determine the part the 'enrich' points to
            # TODO(clairbee): refactor this using the resource locator when available
            self.target_part = config["target"]
            if "project" in config:
                self.target_project = config["project"]
            else:
                self.target_project = None

            pc_logging.debug(
                "Initializing an enrich to %s:%s"
                % (self.target_project, self.target_part)
            )

            # Get the config of the part the 'enrich' points to
            if self.target_project is None:
                augmented_config = project.get_part_config(self.target_part)
            else:
                augmented_config = ctx.get_project(self.target_project).get_part_config(
                    self.target_part
                )
            if augmented_config is None:
                pc_logging.error(
                    "Failed to find the part to enrich: %s" % self.target_part
                )
                return

            augmented_config = copy.deepcopy(augmented_config)
            # TODO(clairbee): ideally whatever we pull from the project is already normalized
            augmented_config = part_config.PartConfiguration.normalize(
                self.target_part, augmented_config
            )

            # Drop fields we don't want to be inherited by enriched clones
            # TODO(clairbee): keep aliases if they are a function of the orignal name
            if "aliases" in augmented_config:
                del augmented_config["aliases"]

            # Fill in all non-enrich-specific properties from the enrich config into
            # the original config
            for prop_to_copy in config:
                if (
                    prop_to_copy == "type"
                    or prop_to_copy == "path"
                    or prop_to_copy == "orig_name"
                    or prop_to_copy == "target"
                    or prop_to_copy == "project"
                    or prop_to_copy == "with"
                ):
                    continue
                augmented_config[prop_to_copy] = config[prop_to_copy]

            # Fill in the parameter values using the simplified "with" option
            if "with" in config:
                for param in config["with"]:
                    augmented_config["parameters"][param]["default"] = config["with"][
                        param
                    ]

            project.init_part_by_config(augmented_config)



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-26
#
# Licensed under Apache License, Version 2.0.
#

import typing

from . import assembly_factory as pf
from . import logging as pc_logging


class AssemblyFactoryAlias(pf.AssemblyFactory):
    target_assembly: str
    target_project: typing.Optional[str]

    def __init__(self, ctx, project, assembly_config):
        with pc_logging.Action("InitAlias", project.name, assembly_config["name"]):
            super().__init__(ctx, project, assembly_config)
            # Complement the config object here if necessary
            self._create(assembly_config)

            self.target_assembly = assembly_config["target"]
            if "project" in assembly_config:
                self.target_project = assembly_config["project"]
            else:
                self.target_project = project.name

            pc_logging.debug(
                "Initializing an alias to %s:%s"
                % (self.target_project, self.target_assembly)
            )

            # Get the config of the assembly the alias points to
            if self.target_project is None:
                self.assembly.desc = "Alias to %s" % self.target_assembly
            else:
                self.assembly.desc = "Alias to %s from %s" % (
                    self.target_assembly,
                    self.target_project,
                )

    def instantiate(self, assembly):
        with pc_logging.Action(
            "Alias", self.project.name, self.assembly_config["name"]
        ):
            self.ctx.stats_assemblies_instantiated += 1

            # TODO(clairbee): resolve the absolute package path?
            target = self.ctx._get_assembly(self.target_assembly, self.target_project)
            target.instantiate(assembly)



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-20
#
# Licensed under Apache License, Version 2.0.

import os
import shutil

from .context import Context
from .assembly import Assembly
from .part import Part
from . import consts

# from . import logging as pc_logging

global _partcad_context
_partcad_context = None


def init(config_path=".") -> Context:
    """Initialize the default context explicitly using the desired path."""
    global _partcad_context
    global _partcad_context_path

    if _partcad_context is None:
        # pc_logging.debug("Initializing (%s)..." % __name__)

        _partcad_context = Context(config_path)
        _partcad_context_path = config_path
    # else:
    #     # The below is useful to troubleshoot common pitfalls.
    #     # But it's not really wrong, and in some cases it's a desired bahavior.
    #     if _partcad_context_path != config_path:
    #         pc_logging.error("Multiple context configurations")
    #         raise Exception("partcad: multiple context configurations")

    return _partcad_context


def fini():
    global _partcad_context
    global _partcad_context_path

    _partcad_context = None
    _partcad_context_path = None


def get_assembly(assembly_name, project_name=consts.THIS, params=None) -> Assembly:
    """Get the assembly from the given project"""
    if project_name is None:
        project_name = consts.THIS
    return init().get_assembly(assembly_name, project_name, params=params)

def get_assembly_cadquery(assembly_name, project_name=consts.THIS, params=None) -> Assembly:
    """Get the assembly from the given project"""
    if project_name is None:
        project_name = consts.THIS
    return init().get_assembly_cadquery(assembly_name, project_name, params=params)

def get_assembly_build123d(assembly_name, project_name=consts.THIS, params=None) -> Assembly:
    """Get the assembly from the given project"""
    if project_name is None:
        project_name = consts.THIS
    return init().get_assembly_build123d(assembly_name, project_name, params=params)


def get_part(part_name, project_name=consts.THIS, params=None) -> Part:
    """Get the part from the given project"""
    if project_name is None:
        project_name = consts.THIS
    return init().get_part(part_name, project_name, params=params)

def get_part_cadquery(part_name, project_name=consts.THIS, params=None) -> Part:
    """Get the part from the given project"""
    if project_name is None:
        project_name = consts.THIS
    return init().get_part_cadquery(part_name, project_name, params=params)

def get_part_build123d(part_name, project_name=consts.THIS, params=None) -> Part:
    """Get the part from the given project"""
    if project_name is None:
        project_name = consts.THIS
    return init().get_part_build123d(part_name, project_name, params=params)


def render(format=None, output_dir=None):
    return init().render(format, output_dir)


def create_package(dst_path=consts.DEFAULT_PACKAGE_CONFIG, private=False):
    if private:
        template_name = "init-private.yaml"
    else:
        template_name = "init-public.yaml"
    src_path = os.path.join(os.path.dirname(__file__), "template", template_name)

    if os.path.exists(dst_path):
        return False
    shutil.copyfile(src_path, dst_path)

    return True



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-22
#
# Licensed under Apache License, Version 2.0.
#

import typing

from . import part_factory as pf
from . import logging as pc_logging


class PartFactoryAlias(pf.PartFactory):
    target_part: str
    target_project: typing.Optional[str]

    def __init__(self, ctx, project, part_config):
        with pc_logging.Action("InitAlias", project.name, part_config["name"]):
            super().__init__(ctx, project, part_config)
            # Complement the config object here if necessary
            self._create(part_config)

            self.target_part = part_config["target"]
            if "project" in part_config:
                self.target_project = part_config["project"]
            else:
                self.target_project = project.name

            pc_logging.debug(
                "Initializing an alias to %s:%s"
                % (self.target_project, self.target_part)
            )

            # Get the config of the part the alias points to
            if self.target_project == "this":
                self.part.desc = "Alias to %s" % self.target_part
            else:
                self.part.desc = "Alias to %s from %s" % (
                    self.target_part,
                    self.target_project,
                )

    def instantiate(self, part):
        with pc_logging.Action("Alias", part.project_name, part.name):
            # TODO(clairbee): resolve the absolute package path?

            target = self.ctx._get_part(self.target_part, self.target_project)
            shape = target.shape
            if not shape is None:
                return shape

            self.ctx.stats_parts_instantiated += 1

            return target.instantiate(part)



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-12-30
#
# Licensed under Apache License, Version 2.0.

import os


class Runtime:
    def __init__(self, ctx, name):
        self.ctx = ctx
        self.name = name
        self.path = os.getenv("HOME", "/tmp") + "/.partcad/runtime/partcad-" + name
        self.initialized = os.path.exists(self.path)



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-12-22
#
# Licensed under Apache License, Version 2.0.
#

DEFAULT_RENDER_WIDTH = 1024
DEFAULT_RENDER_HEIGHT = 1024

DEFAULT_RENDER_SVG_OPTS = {
    "width": DEFAULT_RENDER_WIDTH,
    "height": DEFAULT_RENDER_HEIGHT,
    "showAxes": False,
    "projectionDir": [1, -1, 0.75],
    "strokeWidth": 0.25,
    "strokeColor": [64, 192, 64],
    "hiddenColor": [32, 64, 32],
    "showHidden": False,
    # "marginLeft": 0,
    # "marginTop": 0,
}
